"""Configuration file for the Sphinx documentation builder."""

# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import sys
from pathlib import Path

sys.path.insert(0, str(Path("../..").resolve()))


# -- Project information -----------------------------------------------------

project = "ClopHfit"
copyright = "2021, Daniele Arosio"  # noqa: A001
author = "Daniele Arosio"

# The full version, including alpha/beta/rc tags
release = "0.4.2"

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "autodocsumm",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
    "sphinxcontrib.plantuml",
    "nbsphinx",
    "sphinx_click",
    "myst_parser",
    # 'IPython.sphinxext.ipython_console_highlighting',
]

# Napoleon settings to Default
napoleon_use_ivar = False

autodoc_default_options = {
    "members": True,
    "member-order": "bysource",
    "undoc-members": False,
    "autosummary": True,
}

autodoc_typehints = "description"

# The suffix of source filenames.
source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
}


# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]



"""Test cases for the old scripts."""
from __future__ import annotations

import subprocess
import sys
import typing
from pathlib import Path
from typing import Any
from typing import Iterator
from typing import List
from typing import Tuple

import matplotlib.testing.compare as mpltc  # type: ignore
import pytest
from matplotlib.testing.exceptions import ImageComparisonFailure  # type: ignore

_data = Path(__file__).parent / "data"
tmpoutput = _data / "_tmpoutput"
_expected = _data / "output"

Rscript = Tuple[Tuple[str, str, List[str]], Any]


@pytest.mark.skipif(sys.platform == "win32", reason="does not run on windows")
class TestTitrationFit:
    """Test the old ``fit_titration.py`` script."""

    note_fp = "./NTT-A04-Cl_note"
    csv_files = ["./Meas/A04 Cl_A.csv", "./Meas/A04 Cl_B.csv"]
    res_svd = [
        "K =  17.035\nsK =  0.666\nSA =  -0.0\nsSA =  0.007\nSB =  -0.275\nsSB =  0.002",
        "K =  14.824\nsK =  0.708\nSA =  0.019\nsSA =  0.004\nSB =  -0.274\nsSB =  0.002",
    ]
    res_band = [
        "K =  15.205\nsK =  0.549\nSA =  205506.748\nsSA =  1862.684\nSB =  1781.387\nsSB =  1249.163",
        "K =  15.015\nsK =  0.701\nSA =  253789.494\nsSA =  3055.602\nSB =  -4381.091\nsSB =  2040.904",
    ]

    @pytest.fixture(
        scope="class",
        params=[
            (csv_files[0], res_svd[0], ["-m", "svd"]),
            (csv_files[0], res_band[0], ["-m", "band", "-b", "480", "530"]),
            (csv_files[1], res_svd[1], ["-m", "svd"]),
            (csv_files[1], res_band[1], ["-m", "band", "-b", "480", "530"]),
        ],
    )
    def run_script(self, request: pytest.FixtureRequest) -> Iterator[Rscript]:
        """Run the script as class fixture."""
        cli = Path("../../src/clophfit/old/fit_titration.py")
        csv_file = request.param[0]
        with subprocess.Popen(
            [cli, csv_file, self.note_fp, "-t", "cl", "-d", "_tmpoutput"]
            + request.param[2],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
            cwd=_data,
        ) as process:
            yield request.param, process.communicate()
            if (_data / "bs.txt").exists():
                (_data / "bs.txt").unlink()
            for fp in _expected.glob("*_pdf.png"):
                fp.unlink()

    def test_stdout(self, run_script: Rscript) -> None:
        """It print out results."""
        expected = run_script[0][1]
        assert expected in run_script[1][0]

    @pytest.mark.xfail(reason="Deprecation from dependency.")
    def test_stderr_svd(self, run_script: Rscript) -> None:
        """Test stderr for svd."""
        if run_script[0][2][1] == "svd":
            assert run_script[1][1] == ""
        else:
            raise AssertionError

    def test_stderr_band(self, run_script: Rscript) -> None:
        """Test stderr for band."""
        if run_script[0][2][1] == "band":
            assert run_script[1][1] == ""

    def test_pdf(self, run_script: Rscript) -> None:
        """It saves pdf file."""
        csv_ = run_script[0][0].split("/")[-1].split(".")[0]
        f = "_".join([run_script[0][2][1], csv_, "NTT-A04-Cl_note.pdf"])
        fp_test = tmpoutput / f
        fp_expected = _expected / f
        msg = mpltc.compare_images(fp_test, fp_expected, 80.0)
        if msg:
            raise ImageComparisonFailure(msg)  # pragma: no cover


@pytest.mark.skipif(sys.platform == "win32", reason="does not run on windows")
@typing.no_type_check
class TestTitrationFitGlobal:
    """It test the old ``fit_titration_global.py`` script."""

    dat_files = [
        "./global/pH/D05.dat",
        "./global/Cl/B05-20130628-cor.dat",
    ]
    res = [
        """SA1   683.357   714.804   740.747   767.245   800.925
SB1   246.164   299.394   338.212   374.611   417.296
SA2   11.0487   47.6023   76.2874     104.3   138.128
SB2    491.64   534.936   571.809   611.045   664.602
  K   7.34376   7.51152   7.65166   7.80003   7.99971
bootstrap:""",
        """SA1   26910.8   27072.2   27216.7   27361.3   27522.7
SB1   916.697   1037.35   1144.64    1251.2   1369.35
SA2   3813.97   3970.34   4110.42   4250.52   4406.96
SB2      -INF   24.2754   87.1121   149.801   219.626
  K   7.51099   7.78702   8.03882   8.29508   8.58663
bootstrap:""",
    ]

    @pytest.fixture(
        scope="class",
        params=[
            (dat_files[0], res[0], ["--boot", "3"]),
            (dat_files[1], res[1], ["--boot", "3", "-t", "cl"]),
        ],
    )
    def run_script(self, request: pytest.FixtureRequest) -> Iterator[Rscript]:
        """Run the script as class fixture."""
        cli = "../../src/clophfit/old/fit_titration_global.py"
        dat_file = request.param[0]
        with subprocess.Popen(
            [cli, dat_file, "_tmpoutput"] + request.param[2],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
            cwd=_data,
            shell=False,
        ) as process:
            yield request.param, process.communicate()

    def test_stdout(self, run_script: Rscript) -> None:
        """It print out results."""
        expected = run_script[0][1]
        assert expected in run_script[1][0]

    @pytest.mark.xfail(reason="Deprecation from dependency.")
    def test_stderr(self, run_script: Rscript) -> None:
        """Stderr is empty."""
        assert run_script[1][1] == ""

    @pytest.mark.xfail(reason="Image sizes do not match.")
    def test_png(self, run_script: Rscript) -> None:
        """It saves pdf file."""
        f = ".".join([run_script[0][0], "png"])
        fp_test = tmpoutput / f.rsplit("/", maxsplit=1)[-1]
        fp_expected = _expected / f.lstrip("./")
        msg = mpltc.compare_images(fp_test, fp_expected, 1.01)
        if msg:  # pragma: no cover
            raise ImageComparisonFailure(msg)



"""Binding equations."""
from __future__ import annotations

import numpy as np
from numpy.typing import NDArray


# TODO: use this like fz in prtecan
def kd(
    kd1: float, pka: float, ph: NDArray[np.float_] | float
) -> NDArray[np.float_] | float:
    """Infinite cooperativity model.

    It can describe pH-dependence for chloride dissociation constant.

    Parameters
    ----------
    kd1 : float
        Dissociation constant at pH <= 5.0 (fully protonated).
    pka : float
        Acid dissociation constant.
    ph : Xtype
        pH value(s).

    Returns
    -------
    Xtype
        Predicted Kd value(s).

    Examples
    --------
    >>> kd(10, 8.4, 7.4)
    11.0
    >>> import numpy as np
    >>> kd(10, 8.4, np.array([7.4, 8.4]))
    array([11., 20.])

    """
    return kd1 * (1 + 10 ** (pka - ph)) / 10 ** (pka - ph)


# TODO other from datan



#!/usr/bin/env python

"""Module for global fitting titrations (pH and cl) on 2 datasets
"""
import os
import sys
import argparse
import numpy as np
from lmfit import Parameters, Minimizer, minimize, conf_interval, report_fit
import pandas as pd
import matplotlib.pyplot as plt
# from scipy import optimize


def ci_report(ci):
    """return text of a report for confidence intervals"""
    maxlen = max([len(i) for i in ci])
    buff = []
    add = buff.append
    convp = lambda x: ("%.2f" % (x[0]*100))+'%'
    # I modified "%.5f"
    conv = lambda x: "%.6G" % x[1]
    title_shown = False
    for name, row in ci.items():
        if not title_shown:
            add("".join([''.rjust(maxlen)] +
                        [i.rjust(10) for i in map(convp, row)]))
            title_shown = True
        add("".join([name.rjust(maxlen)] +
                    [i.rjust(10) for i in map(conv,  row)]))
    return '\n'.join(buff)


def residual(pars, x, data=None, titration_type=None):
    """residual function for lmfit
    Parameters
    ----------
    pars: lmfit Parameters()
    x : list of x vectors
    data : list of y vectors

    Return
    ------
    a vector for the residues (yfit - data)
    or the fitted values
    """
    vals = pars.valuesdict()
    SA1 = vals['SA1']
    SB1 = vals['SB1']
    K = vals['K']
    SA2 = vals['SA2']
    SB2 = vals['SB2']
    if titration_type == 'pH':
        model1 = (SB1 + SA1 * 10 ** (K - x[0])) / (1 + 10 ** (K - x[0]))
        model2 = (SB2 + SA2 * 10 ** (K - x[1])) / (1 + 10 ** (K - x[1]))
    elif titration_type == 'cl':
        model1 = (SA1 + SB1 * x[0] / K) / (1 + x[0] / K)
        model2 = (SA2 + SB2 * x[1] / K) / (1 + x[1] / K)
    else:
        print('Error: residual call must indicate a titration type')
        sys.exit()
    if data is None:
        return np.r_[model1, model2]
    return np.r_[model1 - data[0], model2 - data[1]]


def main():
    description = "Fit a pH or Cl titration file: x y1 y2"
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('file',
                        help='the file <x y1 y2> without heads')
    parser.add_argument('out_folder',
                        help='The folder to output the .txt and .png files')
    parser.add_argument('-t', '--titration-of', dest='titration_type',
                        action="store", default="pH", choices=["pH", "cl"],
                        help='Type of titration, pH or cl')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Printout runtime information.increase verbosity')
    parser.add_argument('--boot', dest='nboot', type=int,
                        help='bootstraping using <n> iterations')
    args = parser.parse_args()
    ttype = args.titration_type
    #df = pd.read_csv(args.file, sep=' ', names=['x', 'y1', 'y2'])
    df = pd.read_csv(args.file)
    if not os.path.isdir(args.out_folder):
        os.makedirs(args.out_folder)
    fit_params = Parameters()
    fit_params.add('SA1', value=df.y1[df.x == min(df.x)].values[0], min=0)
    fit_params.add('SB1', value=df.y1[df.x == max(df.x)].values[0], min=0)
    fit_params.add('SA2', value=df.y2[df.x == min(df.x)].values[0], min=0)
    fit_params.add('SB2', value=df.y2[df.x == max(df.x)].values[0], min=0)
    if args.titration_type == "pH":
        fit_params.add('K', value=7, min=4, max=10)
    elif args.titration_type == "cl":
        fit_params.add('K', value=20, min=0, max=1000)
    mini = Minimizer(residual, fit_params, fcn_args=([df.x, df.x],),
                   fcn_kws={'data': [df.y1, df.y2], 'titration_type': ttype})
    res = mini.minimize()
    report_fit(fit_params)
    ci = conf_interval(mini, res, sigmas=[.674, .95])
    print(ci_report(ci))
    # plotting
    xfit = np.linspace(df.x.min(), df.x.max(), 100)
    yfit = residual(fit_params, [xfit, xfit], titration_type=ttype)   # kws={}
    yfit = yfit.reshape(2, len(yfit) // 2)
    plt.plot(df.x, df.y1, 'o', df.x, df.y2, 's', xfit, yfit[0], '-',
             xfit, yfit[1], '-')
    plt.grid(True)
    f_out = os.path.join(args.out_folder, os.path.split(args.file)[1])
    plt.savefig(f_out + ".png")
    if args.nboot:
        bootstrap(df, args.nboot, fit_params, f_out, ttype)


def bootstrap(df, nboot, fit_params, f_out, ttype):
    """Perform bootstrap to estimate parameters variance
    Parameters
    ----------
    df : DataFrame
    nboot : int
    fit_params: lmfit.fit_params
    f_out : string

    Output
    ------
    print results
    plot
    """
    import seaborn as sns
    n_points = len(df)
    kds = []
    sa1 = []
    sb1 = []
    sa2 = []
    sb2 = []
    for i in range(nboot):
        boot_idxs = np.random.randint(0, n_points-1, n_points)
        df2 = df.loc[boot_idxs]
        df2.reset_index(drop=True, inplace=True)
        boot_idxs = np.random.randint(0, n_points-1, n_points)
        df3 = df.loc[boot_idxs]
        df3.reset_index(drop=True, inplace=True)
        try:
            res = minimize(residual, fit_params, args=([df2.x, df3.x],),
                     kws={'data': [df2.y1, df3.y2], 'titration_type': ttype})
            kds.append(res.params['K'].value)
            sa1.append(res.params['SA1'].value)
            sb1.append(res.params['SB1'].value)
            sa2.append(res.params['SA2'].value)
            sb2.append(res.params['SB2'].value)
        except:
            print(df2)
            print(df3)

    dff = pd.DataFrame({'K': kds, 'SA1': sa1, 'SB1': sb1, 'SA2': sa2,
                        'SB2': sb2})
    print("bootstrap: ",
          round(dff.K.quantile(.025), 3),
          round(dff.K.quantile(.163), 3),
          round(dff.K.median(), 3),
          round(dff.K.quantile(.837), 3),
          round(dff.K.quantile(.975), 3))
    sns.set_style('darkgrid')
    g = sns.PairGrid(dff)
    # g.map_diag(sns.kdeplot, lw=3)
    g.map_diag(plt.hist, alpha=0.4)
    g.map_upper(plt.scatter, s=9, alpha=0.6)
    g.map_lower(sns.kdeplot, cmap="Blues_d")
    plt.savefig(f_out + "-bs" + ".png")


if __name__ == '__main__':
    main()



"""Test prtecan module."""
from __future__ import annotations

import re
import sys
from pathlib import Path
from typing import Any

import numpy as np
import pandas as pd
import pytest

from clophfit import prtecan

data_tests = Path(__file__).parent / "Tecan"
pytestmark = pytest.mark.filterwarnings("ignore:OVER")


def test_lookup_listoflines() -> None:
    """It returns indexes for pattern match on col in list of lines."""
    csvl: list[list[str | int | float]] = [
        ["pp", "xy", 1, 2.0],
        ["pp", "xx", 1, 2],
        ["pp", 12, 1, 2],
        ["pp", "yy", 1, 2.0],
        ["a"],
        ["pp", "xy", 1, 2],
    ]
    assert prtecan.lookup_listoflines(csvl, pattern="pp") == [0, 1, 2, 3, 5]
    assert not prtecan.lookup_listoflines(csvl, pattern="xy")  # == []
    assert prtecan.lookup_listoflines(csvl, pattern="xy", col=1) == [0, 5]
    assert prtecan.lookup_listoflines(csvl, pattern="yy", col=1) == [3]
    assert not prtecan.lookup_listoflines(csvl, pattern="yy", col=2)  # == []


def test_strip_lines() -> None:
    """It strips empty fields."""
    lines: list[list[float | int | str]] = [
        ["Excitation Wavelength", "", "", "", 485.0, "nm", "", "", ""]
    ]
    stripped = prtecan.strip_lines(lines)
    assert stripped == [["Excitation Wavelength", 485.0, "nm"]]


def test_extract_metadata() -> None:
    """It extracts metadata correctly."""
    lines: list[list[Any]] = [
        ["Label: Label1", "", "", "", "", "", "", "", "", "", "", "", ""],
        ["Mode", "", "", "", "Fluorescence Top Reading", "", "", "", "", ""],
        ["Shaking (Linear) Amplitude:", "", "", "", 2, "mm", "", "", "", "", ""],
        ["Excitation Wavelength", "", "", "", 400, "nm", "", "unexpected", "", "", ""],
        ["", "Temperature: 26 °C", "", "", "", "", "", "", "", "", ""],
    ]
    expected_metadata = {
        "Label": prtecan.Metadata("Label1"),
        "Mode": prtecan.Metadata("Fluorescence Top Reading"),
        "Shaking (Linear) Amplitude:": prtecan.Metadata(2, ["mm"]),
        "Excitation Wavelength": prtecan.Metadata(400, ["nm", "unexpected"]),
        "Temperature": prtecan.Metadata(26.0, ["°C"]),
    }

    metadata = prtecan.extract_metadata(lines)
    assert metadata == expected_metadata


def test_merge_md() -> None:
    """Merge metadata of both labelblocks and tecanfiles."""
    md1 = {
        "Gain": prtecan.Metadata(93, ["Manual"]),
        "Shaking (Linear) Amplitude:": prtecan.Metadata(2, ["mm"]),
    }
    md2 = {
        "Gain": prtecan.Metadata(93, ["Optimal"]),
        "Shaking (Linear) Amplitude:": prtecan.Metadata(2, ["mm"]),
    }
    mmd = prtecan.merge_md([md1, md2])
    assert mmd["Gain"] == prtecan.Metadata(93)
    assert mmd["Shaking (Linear) Amplitude:"] == prtecan.Metadata(2, ["mm"])


def test_calculate_conc() -> None:
    """Calculates concentration values from Cl additions."""
    additions = [112, 2, 2, 2, 2, 2, 2, 6, 4]
    conc = prtecan.calculate_conc(additions, 1000)
    np.testing.assert_almost_equal(
        conc,
        [0.0, 17.544, 34.483, 50.847, 66.667, 81.967, 96.774, 138.462, 164.179],
        3,
    )


def test_fit_titration() -> None:
    """It fits pH and Cl titrations."""
    x = [3.0, 5, 7, 9, 11.0]
    y = np.array([1.9991, 1.991, 1.5, 1.009, 1.0009])
    fit = prtecan.fit_titration("pH", x, y)
    assert abs(fit["K"][0] - 7) < 0.0000000001
    assert abs(fit["SA"][0] - 2) < 0.0001
    assert abs(fit["SB"][0] - 1) < 0.0001
    x = [0, 5.0, 10, 40, 160, 1000]
    y = np.array([2.0, 1.33333333, 1.0, 0.4, 0.11764706, 0.01980198])
    fit = prtecan.fit_titration("Cl", x, y)
    assert abs(fit["K"][0] - 10) < 0.0000001
    assert abs(fit["SA"][0] - 2) < 0.000000001
    assert abs(fit["SB"][0] - 0) < 0.00000001
    with pytest.raises(NameError, match="kind= pH or Cl"):
        prtecan.fit_titration("unk", x, y)


@pytest.mark.filterwarnings("ignore:OVER")
class TestLabelblock:
    """Test labelblock class."""

    @pytest.fixture(autouse=True)
    def _init(self) -> None:
        """Simulate csvl with 2 labelblocks."""
        csvl = prtecan.read_xls(data_tests / "140220/pH6.5_200214.xls")
        idxs = prtecan.lookup_listoflines(csvl)
        lb0 = prtecan.Labelblock(csvl[idxs[0] : idxs[1]])
        lb1 = prtecan.Labelblock(csvl[idxs[1] :])
        lb0.buffer_wells = ["D01", "D12", "E01", "E12"]
        lb1.buffer_wells = ["D01", "D12", "E01", "E12"]
        # pylint: disable=W0201
        self.lb0 = lb0
        self.lb1 = lb1

    def test_metadata(self) -> None:
        """It parses "Temperature" metadata."""
        assert self.lb0.metadata["Temperature"].value == 25.6
        assert self.lb1.metadata["Temperature"].value == 25.3

    def test_data(self) -> None:
        """It parses data values."""
        assert self.lb0.data["F06"] == 19551
        assert self.lb1.data["H12"] == 543

    def test_data_normalized(self) -> None:
        """Normalize data using some metadata values."""
        assert self.lb0.data_norm["F06"] == pytest.approx(1051.1290323)
        assert self.lb1.data_norm["H12"] == pytest.approx(48.4821429)

    def test_data_buffersubtracted(self) -> None:
        """Calculate buffer value from average of buffer wells and subtract from data."""
        assert self.lb0.buffer == 11889.25
        assert self.lb1.buffer == 56.75
        assert self.lb0.sd_buffer == pytest.approx(450.2490)
        assert self.lb1.sd_buffer == pytest.approx(4.43706)
        assert self.lb0.data_buffersubtracted["F06"] == pytest.approx(7661.75)
        assert self.lb1.data_buffersubtracted["H12"] == pytest.approx(486.25)
        # Can also assign a buffer value.
        self.lb0.buffer = 1
        self.lb1.buffer = 2.9
        assert self.lb0.data_buffersubtracted["F06"] == 19550
        assert self.lb1.data_buffersubtracted["H12"] == 540.1

    def test_data_buffersubtracted_norm(self) -> None:
        """Calculate normalized buffer value from average of buffer wells and subtract from data."""
        assert self.lb0.buffer_norm == pytest.approx(639.20699)
        assert self.lb1.buffer_norm == pytest.approx(5.06696)
        assert self.lb0.sd_buffer_norm == pytest.approx(24.20694)
        assert self.lb1.sd_buffer_norm == pytest.approx(0.396166)
        assert self.lb0.data_buffersubtracted_norm["F06"] == pytest.approx(411.922)
        assert self.lb1.data_buffersubtracted_norm["H12"] == pytest.approx(43.4152)
        # Can also assign a buffer_norm value.
        self.lb0.buffer_norm = 1
        self.lb1.buffer_norm = 0.4821
        assert self.lb0.data_buffersubtracted_norm["F06"] == pytest.approx(1050.13)
        assert self.lb1.data_buffersubtracted_norm["H12"] == pytest.approx(48.0)

    def test___eq__(self) -> None:
        """It is equal to itself. TODO and different from other."""
        assert self.lb0 == self.lb0
        assert self.lb0 != self.lb1
        assert self.lb0.__eq__(1) == NotImplemented

    def test___almost_eq__(self) -> None:
        """It is equal to itself. TODO and different from other."""
        csvl = prtecan.read_xls(data_tests / "290513_7.2.xls")  # Gain=98
        idxs = prtecan.lookup_listoflines(csvl)
        lb11 = prtecan.Labelblock(csvl[idxs[1] :])
        csvl = prtecan.read_xls(data_tests / "290513_8.8.xls")  # Gain=99
        idxs = prtecan.lookup_listoflines(csvl)
        lb12 = prtecan.Labelblock(csvl[idxs[1] :])
        assert lb11 != lb12
        assert lb11.__almost_eq__(lb12)
        assert not lb11.__almost_eq__(self.lb0)

    def test_overvalue(self) -> None:
        """It detects saturated data ("OVER")."""
        csvl = prtecan.read_xls(data_tests / "140220/pH6.5_200214.xls")
        idxs = prtecan.lookup_listoflines(csvl)
        with pytest.warns(
            UserWarning, match=r"OVER\n Overvalue in Label1:A06 of tecanfile "
        ):
            lb = prtecan.Labelblock(csvl[idxs[0] : idxs[1]])
        assert np.nansum(lb.data["A06"]) == np.nansum(np.nan)
        assert np.nansum(lb.data["H02"]) == np.nansum(np.nan)

    def test_raise_missing_column(self) -> None:
        """It raises Exception when a column is missing from the labelblock."""
        csvl = prtecan.read_xls(data_tests / "exceptions/88wells_290212_20.xlsx")
        idxs = prtecan.lookup_listoflines(csvl)
        with pytest.raises(ValueError, match=r"Cannot build Labelblock: not 96 wells?"):
            prtecan.Labelblock(csvl[idxs[0] : len(csvl)])

    def test_raise_missing_row(self) -> None:
        """It raises Exception when a row is missing from the labelblock."""
        csvl = prtecan.read_xls(data_tests / "exceptions/84wells_290212_20.xlsx")
        idxs = prtecan.lookup_listoflines(csvl)
        with pytest.raises(
            ValueError, match="Cannot extract data in Labelblock: not 96 wells?"
        ):
            prtecan.Labelblock(csvl[idxs[0] : len(csvl)])


class TestCsvlFunctions:
    """Test TecanFile reading and parsing functions."""

    csvl = prtecan.read_xls(data_tests / "140220/pH8.3_200214.xls")

    def test_read_xls(self) -> None:
        """The test reads the xls file using cls method."""
        assert len(self.csvl) == 74

    def test_lookup_listoflines(self) -> None:
        """It finds Label occurrences using module function."""
        assert prtecan.lookup_listoflines(self.csvl) == [14, 44]


class TestTecanfile:
    """Test TecanFile class."""

    tf = prtecan.Tecanfile(data_tests / "140220/pH8.3_200214.xls")

    def test_path(self) -> None:
        """It reads the file path."""
        assert self.tf.path == data_tests / "140220/pH8.3_200214.xls"

    def test_metadata(self) -> None:
        """It parses the Date."""
        assert self.tf.metadata["Date:"].value == "20/02/2014"

    def test_labelblocks(self) -> None:
        """It parses "Temperature" metadata and cell data from 2 labelblocks."""
        assert self.tf.labelblocks[0].metadata["Temperature"].value == 25.3
        assert self.tf.labelblocks[1].metadata["Temperature"].value == 25.7
        assert self.tf.labelblocks[0].data["A01"] == 17260
        assert self.tf.labelblocks[1].data["H12"] == 4196

    def test___eq__(self) -> None:
        """It is equal to itself. TODO and different from other."""
        assert self.tf == prtecan.Tecanfile(data_tests / "140220/pH8.3_200214.xls")
        tf2 = prtecan.Tecanfile(data_tests / "140220/pH9.1_200214.xls")
        assert tf2 != self.tf

    def test_warn(self) -> None:
        """It warns when labelblocks are repeated in a Tf as it might compromise grouping."""
        with pytest.warns(UserWarning, match="Repeated labelblocks"):
            prtecan.Tecanfile(data_tests / "exceptions/290212_7.67_repeated_lb.xls")

    def test_filenotfound(self) -> None:
        """It raises FileNotFoundError when the file path does not exist."""
        with pytest.raises(FileNotFoundError):
            prtecan.Tecanfile(Path("pinocchio"))

    def test_missing_label(self) -> None:
        """It raises Exception when there is no Label pattern."""
        with pytest.raises(ValueError, match="No Labelblock found."):
            prtecan.Tecanfile(data_tests / "exceptions/0_Labelblocks_290513_5.5.xlsx")


class TestLabelblocksGroup:
    """Test LabelBlocksGroup class."""

    @pytest.fixture(name="tfs", scope="class")
    def fixture_tfs(self) -> list[prtecan.Tecanfile]:  # pragma: no cover
        """Return a list of `Tecan` files."""
        tf1 = prtecan.Tecanfile(data_tests / "290513_5.5.xls")
        tf2 = prtecan.Tecanfile(data_tests / "290513_7.2.xls")
        tf3 = prtecan.Tecanfile(data_tests / "290513_8.8.xls")
        return [tf1, tf2, tf3]

    @pytest.fixture(autouse=True)
    def _init(self, tfs: list[prtecan.Tecanfile]) -> None:  # pragma: no cover
        """Return 2 label-block groups."""
        lbg1 = prtecan.LabelblocksGroup([tfs[0].labelblocks[0], tfs[1].labelblocks[0]])
        lbg2 = prtecan.LabelblocksGroup([tfs[1].labelblocks[1], tfs[2].labelblocks[1]])
        lbg1.buffer_wells = ["C12", "D01", "D12", "E01", "E12", "F01"]
        lbg2.buffer_wells = ["C12", "D01", "D12", "E01", "E12", "F01"]
        # pylint: disable=W0201
        self.lbg0 = lbg1
        self.lbg1 = lbg2

    def test_metadata(self) -> None:
        """Merge only shared metadata."""
        assert self.lbg0.metadata.get("Temperature") is None
        assert self.lbg1.metadata.get("Temperature") is None
        assert self.lbg1.metadata.get("Gain") is None
        assert self.lbg1.labelblocks[0].metadata["Gain"].value == 98
        assert self.lbg1.labelblocks[1].metadata["Gain"].value == 99
        # Common metadata.
        assert self.lbg0.metadata["Gain"].value == 94
        assert self.lbg0.metadata["Number of Flashes"].value == 10

    def test_data(self) -> None:
        """Merge data."""
        assert self.lbg0.data["A01"] == [18713, 17088]  # type: ignore
        assert self.lbg0.data["H12"] == [28596, 25771]  # type: ignore
        assert self.lbg1.data is None

    def test_data_normalized(self) -> None:
        """Merge data_normalized."""
        np.testing.assert_almost_equal(
            self.lbg1.data_norm["H12"], [693.980, 714.495], 3
        )
        np.testing.assert_almost_equal(
            self.lbg0.data_norm["A01"], [995.372, 908.936], 3
        )

    def test_data_buffersubtracted(self) -> None:
        """Merge data_buffersubtracted."""
        np.testing.assert_almost_equal(
            self.lbg0.data_buffersubtracted["B07"], [7069, 5716.7], 1  # type: ignore
        )
        assert self.lbg1.data_buffersubtracted is None

    def test_data_buffersubtracted_norm(self) -> None:
        """Merge data_buffersubtracted."""
        np.testing.assert_almost_equal(
            self.lbg0.data_buffersubtracted_norm["B07"], [376.01, 304.08], 2
        )
        np.testing.assert_almost_equal(
            self.lbg1.data_buffersubtracted_norm["B07"], [355.16, 348.57], 2
        )

    def test_notequal_labelblocks(self, tfs: list[prtecan.Tecanfile]) -> None:
        """It raises Exception when concatenating unequal labelblocks."""
        with pytest.raises(ValueError, match="Creation of labelblock group failed."):
            prtecan.LabelblocksGroup([tfs[1].labelblocks[0], tfs[2].labelblocks[1]])


class TestTecanfileGroup:
    """Group tecanfiles properly."""

    class TestAllEqLbgs:
        """Test TfG with 2 LbG in the same order."""

        @pytest.fixture(name="tfg", scope="class")
        def fixture_tfg(self) -> prtecan.TecanfilesGroup:  # pragma: no cover
            """Initialize Tfg with 2 LbG in the same order."""
            filenames = ["290513_5.5.xls", "290513_7.2.xls"]
            tecanfiles = [prtecan.Tecanfile(data_tests / f) for f in filenames]
            return prtecan.TecanfilesGroup(tecanfiles)

        def test_metadata(self, tfg: prtecan.TecanfilesGroup) -> None:
            """Parse general metadata."""
            assert tfg.metadata["Plate"].value == "PE 96 Flat Bottom White   [PE.pdfx]"
            assert tfg.metadata["System"].value == "TECANROBOT"

        def test_labelblocksgroups(self, tfg: prtecan.TecanfilesGroup) -> None:
            """Generate 2 LbG with .data and .metadata."""
            lbg0 = tfg.labelblocksgroups[0]
            lbg1 = tfg.labelblocksgroups[1]
            # metadata
            assert lbg0.metadata["Number of Flashes"].value == 10.0
            assert lbg1.metadata["Gain"].value == 98.0
            # data normalized ... enough in lbg
            # data
            assert lbg0.data["A01"] == [18713, 17088]  # type: ignore
            assert lbg0.data["H12"] == [28596, 25771]  # type: ignore
            assert lbg1.data["A01"] == [7878, 8761]  # type: ignore
            assert lbg1.data["H12"] == [14226, 13602]  # type: ignore

    class TestAlmostEqLbgs:
        """Test TfG when 1 LbG equal and a second with almost equal labelblocks."""

        @pytest.fixture(autouse=True)
        def _init(self) -> None:  # pragma: no cover
            """Initialize Tfg when 1 LbG equal and a second with almost equal labelblocks."""
            filenames = [
                "290513_5.5.xls",  # Label1 and Label2
                "290513_7.2.xls",  # Label1 and Label2
                "290513_8.8.xls",  # Label1 and Label2 with different metadata
            ]
            tecanfiles = [prtecan.Tecanfile(data_tests / f) for f in filenames]
            # pylint: disable=W0201
            with pytest.warns(UserWarning) as self.record:
                self.group = prtecan.TecanfilesGroup(tecanfiles)

        def test_warn(self) -> None:
            """Warn about labelblocks anomaly."""
            assert "Different LabelblocksGroup among filenames:" in str(
                self.record[0].message
            )

        def test_labelblocksgroups(self) -> None:
            """Generate 1 LbG with .data and .metadata."""
            lbg0 = self.group.labelblocksgroups[0]
            # metadata
            assert lbg0.metadata["Number of Flashes"].value == 10.0
            assert lbg0.metadata["Gain"].value == 94
            # data
            assert lbg0.data["A01"] == [18713.0, 17088.0, 17123.0]  # type: ignore
            assert lbg0.data["H12"] == [28596.0, 25771.0, 28309.0]  # type: ignore

        def test_mergeable_labelblocksgroups(self) -> None:
            """Generate 1 Lbg only with .data_normalized and only common .metadata."""
            lbg1 = self.group.labelblocksgroups[1]
            # metadata
            assert lbg1.metadata["Number of Flashes"].value == 10.0
            assert lbg1.metadata.get("Gain") is None
            assert lbg1.data is None
            # data_normalized
            np.testing.assert_almost_equal(
                lbg1.data_norm["A01"], [401.9387755, 446.9897959, 450.0]
            )
            np.testing.assert_almost_equal(
                lbg1.data_norm["H12"], [725.8163265, 693.9795918, 714.4949494]
            )

    class TestOnly1commonLbg:
        """Test TfG with different number of labelblocks, but mergeable."""

        @pytest.fixture(autouse=True)
        def _init(self) -> None:
            """Initialize Tfg with different number of labelblocks, but mergeable."""
            filenames = [
                "290212_5.78.xls",  # Label1 and Label2
                "290212_20.xls",  # Label2 only
                "290212_100.xls",  # Label2 only
            ]
            tecanfiles = [prtecan.Tecanfile(data_tests / f) for f in filenames]
            # pylint: disable=W0201
            with pytest.warns(UserWarning) as self.record:
                self.group = prtecan.TecanfilesGroup(tecanfiles)

        def test_warn(self) -> None:
            """Warn about labelblocks anomaly."""
            assert "Different LabelblocksGroup among filenames" in str(
                self.record[0].message
            )

        def test_labelblocksgroups(self) -> None:
            """Generates 1 LbG with .data and .metadata."""
            lbg = self.group.labelblocksgroups[0]
            # metadata
            assert lbg.metadata["Number of Flashes"].value == 10.0
            assert lbg.metadata["Gain"].value == 93.0
            # data
            assert lbg.data["A01"] == [6289, 6462, 6465]  # type: ignore
            assert lbg.data["H12"] == [4477, 4705, 4918]  # type: ignore

    class TestFailToMerge:
        """Test TfG without mergeable labelblocks."""

        def test_raise_exception(self) -> None:
            """Raise Exception when there is no way to build labelblocksGroup."""
            filenames = ["290513_5.5.xls", "290513_5.5_bad.xls"]
            tecanfiles = [prtecan.Tecanfile(data_tests / f) for f in filenames]
            with pytest.raises(
                ValueError,
                match=r"No common labelblock in filenames: .*290513_5.5_bad.xls",
            ):
                prtecan.TecanfilesGroup(tecanfiles)


class TestTitration:
    """Test Titration class."""

    tit_ph = prtecan.Titration.fromlistfile(data_tests / "list.pH")
    tit_cl = prtecan.Titration.fromlistfile(data_tests / "list.cl20")

    @pytest.mark.filterwarnings("ignore: Different LabelblocksGroup")
    def test_conc(self) -> None:
        """It reads pH values."""
        assert self.tit_ph.conc == [5.78, 6.38, 6.83, 7.24, 7.67, 8.23, 8.82, 9.31]

    def test_labelblocksgroups(self) -> None:
        """It reads labelblocksgroups data and metadata."""
        lbg0 = self.tit_ph.labelblocksgroups[0]
        lbg1 = self.tit_ph.labelblocksgroups[1]
        # metadata
        assert lbg0.metadata["Number of Flashes"].value == 10.0
        # pH9.3 is 93 Optimal not Manual
        assert lbg1.metadata["Gain"] == prtecan.Metadata(93.0)
        # data
        assert lbg0.data["A01"][::2] == [30344, 31010, 33731, 37967]  # type: ignore
        assert lbg1.data["A01"][1::2] == [9165, 15591, 20788, 22534]  # type: ignore
        assert lbg0.data["H12"][1::2] == [20888, 21711, 23397, 25045]  # type: ignore
        assert lbg1.data["H12"] == [4477, 5849, 7165, 8080, 8477, 8822, 9338, 9303]  # type: ignore

    def test_labelblocksgroups_cl(self) -> None:
        """It reads labelblocksgroups data for Cl too."""
        lbg = self.tit_cl.labelblocksgroups[0]
        assert lbg.data["A01"] == [6462, 6390, 6465, 6774]  # type: ignore
        assert lbg.data["H12"] == [4705, 4850, 4918, 5007]  # type: ignore

    def test_export_data(self, tmp_path: Path) -> None:
        """It exports titrations data to files e.g. "A01.dat"."""
        self.tit_ph.export_data(tmp_path)
        a01 = pd.read_csv(tmp_path / "dat/A01.dat")
        h12 = pd.read_csv(tmp_path / "dat/H12.dat")
        assert a01["y1"].tolist()[1::2] == [30072, 32678, 36506, 37725]
        assert a01["y2"].tolist()[1::2] == [9165, 15591, 20788, 22534]
        assert h12["y1"].tolist()[1::2] == [20888, 21711, 23397, 25045]
        assert h12["y2"].tolist() == [4477, 5849, 7165, 8080, 8477, 8822, 9338, 9303]

    def test_raise_listfilenotfound(self) -> None:
        """It raises FileNotFoundError when list.xx file does not exist."""
        with pytest.raises(FileNotFoundError, match="Cannot find: aax"):
            prtecan.Titration.fromlistfile(Path("aax"))

    def test_bad_listfile(self) -> None:
        """It raises Exception when list.xx file is ill-shaped."""
        with pytest.raises(ValueError, match=r"Check format .* for listfile: .*"):
            prtecan.Titration.fromlistfile(data_tests / "list.pH2")


@pytest.mark.filterwarnings("ignore:OVER")
class TestTitrationAnalysis:
    """Test TitrationAnalysis class."""

    @pytest.fixture(autouse=True)
    def _init(self) -> None:
        """Set up test class."""
        # Use fixture to capture UserWarning "OVER"
        # pylint: disable=W0201
        self.titan = prtecan.TitrationAnalysis.fromlistfile(
            data_tests / "140220/list.pH"
        )
        self.titan.load_additions(data_tests / "140220/additions.pH")
        self.titan.load_scheme(data_tests / "140220/scheme.txt")

    def test_scheme(self) -> None:
        """It finds well position for buffer samples."""
        assert self.titan.scheme.buffer == ["D01", "E01", "D12", "E12"]

    def test_raise_listfilenotfound(self) -> None:
        """It raises OSError when scheme file does not exist."""
        with pytest.raises(
            FileNotFoundError, match=r"No such file or directory: 'aax'"
        ):
            self.titan.load_scheme(Path("aax"))

    def test_raise_listfile_exception(self) -> None:
        """It raises AssertionError when scheme.txt file is ill-shaped."""
        bad_schemefile = data_tests / "140220/scheme0.txt"
        msg = f"Check format [well sample] for schemefile: {bad_schemefile}"
        with pytest.raises(ValueError, match=re.escape(msg)):
            self.titan.load_scheme(bad_schemefile)

    def test_subtract_bg(self) -> None:
        """It subtracts buffer average values."""
        lbg0 = self.titan.labelblocksgroups[0]
        lbg1 = self.titan.labelblocksgroups[1]
        np.testing.assert_almost_equal(
            lbg0.data_norm["E01"][::2],
            [601.72, 641.505, 674.355, 706.774],
            3,
        )
        if lbg0.data:
            assert lbg0.data["E01"][::2] == [11192.0, 11932.0, 12543.0, 13146.0]
        if lbg0.data_buffersubtracted:
            np.testing.assert_array_equal(
                lbg0.data_buffersubtracted["A12"][::3],
                [8084.5, 16621.75, 13775.0],
            )
        np.testing.assert_array_equal(
            lbg1.data_buffersubtracted["A12"][::3],  # type: ignore
            [9758.25, 1334.0, 283.5],
        )

    def test_dilution_correction(self) -> None:
        """It applies dilution correction read from file listing additions."""
        np.testing.assert_array_equal(self.titan.additions, [100, 2, 2, 2, 2, 2, 2])  # type: ignore
        np.testing.assert_almost_equal(
            self.titan.data_dilutioncorrected[1]["A12"],  # type: ignore
            [9758.25, 7524.795, 3079.18, 1414.04, 641.79, 402.325, 317.52],
        )

    def test_data_dilutioncorrected_norma(self) -> None:
        """It normalizes data."""
        np.testing.assert_almost_equal(
            self.titan.data_dilutioncorrected_norm[0]["A12"][::2],  # type: ignore
            [434.65, 878.73, 975.58, 829.46],
            2,
        )
        np.testing.assert_almost_equal(
            self.titan.data_dilutioncorrected_norm[1]["A12"][::2],  # type: ignore
            [871.272, 274.927, 57.303, 28.35],
            3,
        )

    def test_keys(self) -> None:
        """It gets well positions for ctrl and unknown samples."""
        assert set(self.titan.scheme.names) == {"NTT", "G03", "V224Q", "S202N"}
        x = {"B12", "H12", "F01", "C12", "F12", "C01", "H01", "G12", "B01", "G01"}
        assert set(self.titan.scheme.ctrl) - {"A01", "A12"} == x

    @pytest.mark.skipif(sys.platform == "win32", reason="broken on windows")
    def test_fit(self) -> None:
        """It fits each label separately."""
        self.titan.fit("pH", nrm=True, bg=True, dil=True)
        fit0 = self.titan.fittings[0].sort_index()
        fit1 = self.titan.fittings[1].sort_index()
        df0 = pd.read_csv(data_tests / "140220/fit0.csv", index_col=0)
        df1 = pd.read_csv(data_tests / "140220/fit1.csv", index_col=0)
        pd.testing.assert_frame_equal(df0.sort_index(), fit0, rtol=1e0)
        pd.testing.assert_frame_equal(
            df1,
            fit1,
            check_like=True,
            check_categorical=False,
            atol=1e-3,
        )
        # fit up to the second-last data point
        self.titan.fit("pH", fin=-1, nrm=True, bg=True, dil=True)
        fit0 = self.titan.fittings[0].sort_index()
        fit1 = self.titan.fittings[1].sort_index()
        df0 = pd.read_csv(data_tests / "140220/fit0-1.csv", index_col=0)
        df1 = pd.read_csv(data_tests / "140220/fit1-1.csv", index_col=0)
        pd.testing.assert_frame_equal(
            df0.sort_index()[1:], fit0[1:], check_like=True, atol=1e-4
        )
        pd.testing.assert_frame_equal(df1.sort_index(), fit1, atol=1e-5)



"""Test ``clop`` cli."""
from __future__ import annotations

from pathlib import Path

import pytest
from click.testing import CliRunner

from clophfit.__main__ import clop


def test_eq1() -> None:
    """It runs XXX pr.tecan and generates correct results and graphs."""
    runner = CliRunner()
    result = runner.invoke(clop, ["eq1", "2", "2", "2"])
    assert result.exit_code == 0
    assert "4." in result.output


@pytest.mark.filterwarnings("ignore:OVER")
def test_prtecan(tmp_path: Path) -> None:
    """It runs XXX pr.tecan and generates correct results and graphs."""
    out = tmp_path / "out3"
    out.mkdir()
    runner = CliRunner()
    result = runner.invoke(
        clop,
        [
            "prtecan",
            "tests/Tecan/140220/list.pH",
            "--out",
            str(out),
            "--fit",
            "--scheme",
            "tests/Tecan/140220/scheme.txt",
            "--bg",
        ],
    )
    assert result.exit_code == 0



#!/usr/bin/python
#

import argparse
import os
import sys
import pandas as pd


corr = {'pH': [1, 1.02, 1.04, 1.06, 1.08, 1.1, 1.12],
        'cl': [1.000000, 1.017857, 1.035714, 1.053571, 1.071429, 1.089286,
               1.107143, 1.160714, 1.196429]}

corr = {'pH': [1, 1.02, 1.04, 1.06, 1.08, 1.1],
        'cl': [1.000000, 1.017857, 1.035714, 1.053571, 1.071429, 1.089286,
               1.107143, 1.160714, 1.196429]}


def main():
    description = "make correction for dilution"
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('scheme',
                        help="scheme.txt file")
    args = parser.parse_args()

    if not (os.path.exists('./pH') and os.path.exists('./cl')):
        print('must be in a meas folder after pr.tecan run list.pH and list.cl')
        sys.exit(1)

    scheme = pd.read_table(args.scheme)
    l = list(scheme.well[scheme.sample == 'buffer'])

    df = pd.read_csv('pH/' + l[0] + '.dat')
    for f in l[1:]:
        df = df.append(pd.read_csv('pH/' + f + '.dat'))
    print('pH')
    print('y1 :', df.y1.mean(), df.y1.std())
    print('y2 :', df.y2.mean(), df.y2.std())

    df = pd.read_csv('cl/' + l[0] + '.dat')
    for f in l[1:]:
        df = df.append(pd.read_csv('cl/' + f + '.dat'))
    print('cl')
    print('y1 :', df.y1.mean(), df.y1.std())
    print('y2 :', df.y2.mean(), df.y2.std())


main()
